
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "png.h"


unsigned int rusty_filter[32768];


int gbc_get_value(int min, int max, int v) {

  return min+((float)(max-min))*(2.0f*(v/31.0f)-(v/31.0f)*(v/31.0f));
}


void gbc_gen_rusty_filter(void) {

  int r, g, b, nr, ng, nb;


  for (r = 0; r < 32; r++) {
    for (g = 0; g < 32; g++) {
      for (b = 0; b < 32; b++) {
        nr = gbc_get_value(gbc_get_value(4, 14, g), gbc_get_value(24, 29, g), r) - 4;
        ng = gbc_get_value(gbc_get_value(4 + gbc_get_value(0, 5, r), 14 + gbc_get_value(0, 3, r), b),
                           gbc_get_value(24 + gbc_get_value(0, 3, r), 29 + gbc_get_value(0, 1, r), b), g) - 4;
        nb = gbc_get_value(gbc_get_value(4 + gbc_get_value(0, 5, r), 14 + gbc_get_value(0, 3, r), g),
                           gbc_get_value(24 + gbc_get_value(0, 3, r), 29 + gbc_get_value(0, 1, r), g), b) - 4;
        rusty_filter[(b << 10) | (g << 5) | r] = (nb << 10) | (ng << 5) | nr;
      }
    }
  }
}


int main(int argc, char *argv[]) {

  unsigned char *d, *t, *s, co[4*3];
  int dx, dy, type, i, j, k, n, x, y, c, pa[4], rusty = NO;
  float dis, dit, dir, dig, dib;
  char name[256];
  FILE *f;


  /* parse the arguments */
  i = FAILED;
  if (argc == 4) {
    i = SUCCEEDED;
    if (strcmp("-r", argv[1]) == 0)
      rusty = YES;
    else if (strcmp("-n", argv[1]) == 0)
      rusty = NO;
    else
      i = FAILED;
  }

  if (i == FAILED) {
    fprintf(stderr, "USAGE: %s -[rn] <PNG FILE> <OUT FILE>\n", argv[0]);
    return -1;
  }

  /* generate the palette table */
  if (rusty == YES)
    gbc_gen_rusty_filter();

  /* load the image */
  if (png_load(argv[2], &dx, &dy, &type, &d) == FAILED)
    return -1;

  if (type == GL_RGB)
    n = 3;
  else
    n = 4;

  /* write out the scanlines for dithering */
  for (y = 0; y < 18*4; y++) {
    sprintf(name, "tmp_%.2d.png", y);
    png_save(name, 80, 1, type, &d[y*dx*n]);
  }
  free(d);

  /* dither the scanlines */
  for (y = 0; y < 18*4; y++) {
    sprintf(name, "convert -dither -colors 4 tmp_%.2d.png tmp_%.2d.png", y, y);
    system(name);
  }

  t = malloc((20+4*2)*(18*4));
  if (t == NULL) {
    fprintf(stderr, "MAIN: Out of memory error.\n");
    return -1;
  }

  f = fopen(argv[3], "wb");
  if (f == NULL) {
    fprintf(stderr, "MAIN: Could not open file \"%s\" for writing.\n", argv[2]);
    free(t);
    return 1;
  }

  /* read the scanlines */
  for (i = 0, y = 0; y < 18*4; y++) {
    sprintf(name, "tmp_%.2d.png", y);
    if (png_load(name, &i, &j, &type, &s) == FAILED)
      return 1;

    /* remove the scanline file */
    remove(name);

    if (type == GL_RGB)
      n = 3;
    else
      n = 4;

    /* find the unique colors */
    for (c = 0, j = 0, x = 0; x < 80; x++) {
      for (k = 0; k < c; k++) {
	if (co[k*3 + 0] == s[x*n + 0] && co[k*3 + 1] == s[x*n + 1] && co[k*3 + 2] == s[x*n + 2])
	  break;
      }
      if (k == c) {
	/* found a new color */
	if (c == 4)
	  fprintf(stderr, "MAIN: Four color limit exceeded in \"%s\"!\n", name);
	co[c*3 + 0] = s[x*n + 0];
	co[c*3 + 1] = s[x*n + 1];
	co[c*3 + 2] = s[x*n + 2];
	if (c < 4)
	  c++;
      }
    }
    /* DEBUG
    fprintf(stderr, "%d %d %d, %d %d %d, %d %d %d, %d %d %d\n",
	    co[0], co[1], co[2], co[3],
	    co[4], co[5], co[6], co[7],
	    co[8], co[9], co[10], co[11]);
    */

    if (rusty == YES) {
      /* convert the colors */
      for (x = 0; x < 4; x++) {
	/* find the best match */
	c = -1;
	dis = 100000000.0f;
	for (k = 0; k < 32768; k++) {
	  dir = ((float)(co[x*3 + 0]>>3)) - (rusty_filter[k]&31);
	  dig = ((float)(co[x*3 + 1]>>3)) - ((rusty_filter[k]>>5)&31);
	  dib = ((float)(co[x*3 + 2]>>3)) - ((rusty_filter[k]>>10)&31);
	  dit = dir*dir + dig*dig + dib*dib;
	  if (dit < dis) {
	    dis = dit;
	    c = k;
	  }
	}
	pa[x] = c;
      }
    }

    /* write the pixel data */
    x = 0;
    while (x < 80) {
      i = 0;
      for (c = 0; c < 4; c++) {
	if (co[c*3 + 0] == s[x*n + 0] && co[c*3 + 1] == s[x*n + 1] && co[c*3 + 2] == s[x*n + 2])
	  break;
      }
      x++;
      i |= c<<6;
      for (c = 0; c < 4; c++) {
	if (co[c*3 + 0] == s[x*n + 0] && co[c*3 + 1] == s[x*n + 1] && co[c*3 + 2] == s[x*n + 2])
	  break;
      }
      x++;
      i |= c<<4;
      for (c = 0; c < 4; c++) {
	if (co[c*3 + 0] == s[x*n + 0] && co[c*3 + 1] == s[x*n + 1] && co[c*3 + 2] == s[x*n + 2])
	  break;
      }
      x++;
      i |= c<<2;
      for (c = 0; c < 4; c++) {
	if (co[c*3 + 0] == s[x*n + 0] && co[c*3 + 1] == s[x*n + 1] && co[c*3 + 2] == s[x*n + 2])
	  break;
      }
      x++;
      i |= c;
      fprintf(f, "%c", i);
    }

    /* write the palette */
    if (rusty == NO) {
      /* 1:1 R5G5B5 values, no color correction */
      fprintf(f, "%c%c", (co[0]>>3) | (((co[1]>>3)&7)<<5), (co[1]>>6) | ((co[2]>>3)<<2));
      fprintf(f, "%c%c", (co[3]>>3) | (((co[4]>>3)&7)<<5), (co[4]>>6) | ((co[5]>>3)<<2));
      fprintf(f, "%c%c", (co[6]>>3) | (((co[7]>>3)&7)<<5), (co[7]>>6) | ((co[8]>>3)<<2));
      fprintf(f, "%c%c", (co[9]>>3) | (((co[10]>>3)&7)<<5), (co[10]>>6) | ((co[11]>>3)<<2));
    }
    else {
      /* use the reverse mapped colors */
      fprintf(f, "%c%c", pa[0]&255, (pa[0]>>8)&255);
      fprintf(f, "%c%c", pa[1]&255, (pa[1]>>8)&255);
      fprintf(f, "%c%c", pa[2]&255, (pa[2]>>8)&255);
      fprintf(f, "%c%c", pa[3]&255, (pa[3]>>8)&255);
    }

    free(s);
  }

  fclose(f);

  return 0;
}
